"""
Pelecanus: an application for navigating and editing
nested JSON.

Named 'pelecanus' after Pelecanus occidentalis, the
brown Pelican of California and the Eastern Pacific:

http://www.nps.gov/chis/naturescience/brown-pelican.htm

which is a wonderful bird and which deserves to have something
named after it and because I got tired of writing "NestedJson".
"""
from .pelicanjson import PelicanJson  # noqa

__version__ = '0.5.2'
